#! /usr/bin/env python
# -*- coding: utf-8 -*-
################################################################################
# Python API & Sample for miDoor
# Copyright (C) 2014 by Mission Machines, Inc.  All rights reserved.
# 
# http://www.missionmachines.com
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met: 
#
# - Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimer. 
# - Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution. 
# - Neither name of Mission Machines, Inc. or the names of its contributors 
#   may be used to endorse or promote products derived from this software 
#   without specific prior written permission.
# 
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MISSION MACHINES, Inc,
#   OR * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
#   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
#   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
#   OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#	Questions: andrew@missionmachines.com or support@missionmachines.com
#
from __future__ import with_statement

__VERSION__ = "0.1"

from contextlib import contextmanager
import socket
import re
import sys

#__stdssl__ is a flag to indicate to use the 'ssl' module of Python 2.7+.
# 			True means to use the 'ssl' module.  False means to use OpenSSL.
#			OpenSSL is needed for old versions of Python such as 2.5.
#			Mac includes Python OpenSSL.  If using OpenSSL on other platforms
#			you will have to make sure it is installed on that platform.
__stdssl__ = True
if __stdssl__:
	try:
		import ssl
	except ImportError:
		__stdssl__ = False
	
import OpenSSL.SSL
import OpenSSL.crypto
import struct

def mi_enum(**enums):
    return type('Enum', (object,), enums)
 
# Door Status
DoorStatus = mi_enum(Unknown=-1, Closed=0, Opened=1)
# Alarm Trigger
AlarmTrigger = mi_enum(NoAlarm=-1, Reserved=0, OpenByUser=1, OpenTooLong=2, OpenTime=3, 
					   DoorChange=4)

class miDoor:
	
	def __init__(self, ipAddress, user, password, port=2553):
		self._Socket = None
		self.reset()
		self._IPAddress = ipAddress
		self._Username = user
		self._Password = password
		self._Port = port
	
	#
	#	"Private" Methods
	#
	
	# OpenSSL Support Routines
	#	These are used only if __stdssl__ is False
	#
	# Sexy way of ignoring unwanted exceptions
	@contextmanager
	def _magic(self):
		"""Private method.  Perform an SSL operation and ignore non-error exceptions. """
		try:
			yield
		except (OpenSSL.SSL.WantReadError, OpenSSL.SSL.WantWriteError, OpenSSL.SSL.WantX509LookupError):
			pass

	#
	# Do an OpenSSL connection
	def _SSLConnect(self, addr):
		"""Private method.  Create OpenSSL connection"""
		self._SSLConnection = None
		self._SSLContext = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv3_METHOD)
		self._SSLConnection = OpenSSL.SSL.Connection(self._SSLContext, self._Socket)
		self._SSLConnection.connect(addr)
		self._SSLConnection.setblocking(1)
		s = long(self._Timeout/1000)
		us = long((self._Timeout-(s*1000))*1000)
		tv = struct.pack('ll', s, us)
		self._SSLConnection.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, tv)
		with self._magic():
			self._SSLConnection.do_handshake()	
			
	# Send a message to a miDoor device
	# Returns True if message sent
	def _Send(self, msg):
		"""Private method.  Send text to device"""
		if __stdssl__ == True:
			try:
				self._Socket.send(msg)
			except:
				return False
			return True
		else:
			try:
				self._SSLConnection.send(msg)
			except:
				print "Write Exception: " + str(sys.exc_info()[0]), sys.exc_info()
				return False
			return True
			
	# Receive data from the miDoor device within _Timeout milliseconds.
	# Wait for newline character to complete returned data
	# String or None if there were an error or no data
	def _Read(self):
		"""Private method.  Receive line of text from device"""
		data = None
		if __stdssl__ == True:
			try:
				data =  self._Socket.recv(512)
			except:
				return None
		else:
			try:
				with self._magic():
					data = self._SSLConnection.recv(512)
			except:
				return None
		if data != None and len(data) == 0:
			return None
		return data

	# Quick check that socket is active and alive
	def _CheckConnection(self):
		"""Private method.  See if network connection valid"""
		return self._Socket != None
		
	# Create an array of values using white space and ':' as separators
	# if noSplitAfter is non-zero then at noSplitAfter index take rest of message
	# string as one single string (do not split it.)  If noSplitAfter is used then
	# trailing whitespace (i.e. \n) are removed
	# Array of parsed text or None for error
	def _SplitString(self, msg, noSplitAfter=0):
		"""Private method.  'Parse' message into token array"""
		res = re.split('[\s:]', msg, noSplitAfter)
		cnt = res.count('')
		while cnt > 0:
			res.remove('')
			cnt = cnt - 1
		if noSplitAfter > 0:
			res[len(res)-1] = res[len(res)-1].rstrip()
		return res
	
	# Check if a returned message indicate error
	# True if message is error message, False if valid response
	def _IsError(self, msg):
		"""Private method.  Determines if message is an error message."""
    	#STATUS ERROR
    	#0      1
		res = self._SplitString(msg)
		if len(res) < 2:
			return False
		if res[0].upper() == 'STATUS' and res[1].upper() == 'ERROR':
			return True
		return False
    
    # Query the device for current status and update local variables
    # _LastState, _LastAge, _LastTemperature, _LastAlarm are all updated on valid
    # response.  Device must be actively connected.
    # If there is an active alarm then _AlarmTrigger and _AlarmMessage are updated
    # True if status queried and variables updated, False otherwise
	def _UpdateDoorStatus(self):
		"""Private method.  Query device for door status and parse into local vars"""
		#STATE <port>: <state #> <last ms> <changed ms> <state text> <celsius> <ALARM:CALM> <active connections> <switch setting: MT/SW/G1> <admin: AD/NO>
		#0      1       2         3         4            5            6         7            8                    9                          10   
		if self._CheckConnection() == False:
			return False
		if self._Send('GET\n') == False:
			return False
		msg = self._Read()
		if msg == None:
			return False
		if self._IsError(msg):
			return False
		res = self._SplitString(msg)
		if res[0].upper() != 'STATE':
			return False
		val = int(res[2])
		self._LastState = val
		self._LastAge = long(res[4])
		self._LastTemperature = float(res[6])/10.0
		self._LastAlarm = int(res[7])
		if self._LastAlarm > AlarmTrigger.Reserved:
			if self._UpdateAlarmStatus() == False:
				return False
		return True

	# Queries the device for alarm status
	# Updated _AlarmTrigger & _AlarmMessage.
	# Returns True if successful query False otherwise
	def _UpdateAlarmStatus(self):
		"""Private method.  Query active alarm details"""
		#ALARM <instance #>: <alarm trigger> <alarm message>
		#0      1             2               3
		if self._CheckConnection() == False:
			return False
		if self._Send('GET ALARM\n') == False:
			return False
		msg = self._Read()
		if msg == None:
			return False
		if self._IsError(msg):
			return False
		res = self._SplitString(msg,4)
		if res[0].upper() != 'ALARM':
			return False		
		self._AlarmTrigger = int(res[2])
		self._AlarmMessage = res[3]
		return True
	
	# Read event at specified index into an array.  Indexes start at 0 for the 
	# most recent event and work their way back.
	# Array returned is formatted as:	
	#	[eventId, DoorStatus, user, uptime, utc, temp, AlarmTrigger]
	# None is returned on error or if no more events
	def _ReadEvent(self, index):
		"""Private method.  Reads event into array at index"""
		#EVENT 0: <event id> <state> <state text> <user> <uptime> <utc> <temp> <alarm>
		#0     1   2          3       4            5      6        7     8      9
		if self._CheckConnection() == False:
			return None
		if self._Send('NEXT 0: LOG '+str(index)+'\n') == False:
			return None
		msg = self._Read()
		if msg == None:
			return None
		if self._IsError(msg):
			return None
		res = self._SplitString(msg)
		#Parse event msg into array
		event = []
		if res[0].upper() != 'EVENT':
			return None	
		event.append(long(res[2]))
		event.append(int(res[3]))
		event.append(res[5])
		event.append(long(res[6]))
		event.append(long(res[7]))
		event.append(float(res[8])/10.0)
		event.append(int(res[9]))
		return event
	
	# Read events form device and load into array
	# Array is ordered from newest to oldest so that newest event is at index 0 
	# and the oldest event at the end of the array
	# Up to maxEvents will be read.  Can be less than maxEvents
	# If there are no events then None is returned
	def _ReadAllEvents(self, maxEvents):
		"""Private Method.  Read up to the most 'maxEvents' recent events"""
		eventList = []
		index = 0
		while True:
			maxEvents = maxEvents - 1
			if maxEvents < 0:
				break		
			event = self._ReadEvent(index)
			if event == None:
				break
			eventList.append(event)
			index = index + 1			
		if len(eventList) == 0:
			return None
		return eventList
	
	# Issue Push simulation
	# Return True if push was accepted, false otherwise
	# If holdTime > 0 then this is the # of MS to hold the button.  This argument
	# not yes supported.  <= 0 causes miDoor to use its default value.
	def _PushButton(self, holdTime=-1):
		"""Private Method.  Issue Push Operation on the door."""
		if self._CheckConnection() == False:
			return False
		if self._Send('SET 0: push\n') == False:
			return False
		msg = self._Read()
		if msg == None:
			return False
		if self._IsError(msg):
			return False
		res = self._SplitString(msg)
		if res[0].upper() != 'STATUS':
			return False
		if res[2].upper() != 'ACCEPTED':
			return False
		return True
		
	# Issue Clear Alarm commands
	# Return True if clear was accepted, false otherwise
	def _ClearAlarm(self):
		"""Private Method.  Clear active alarm."""
		if self._CheckConnection() == False:
			return False
		if self._Send('SET 0: CLEAR ALARM\n') == False:
			return False
		msg = self._Read()
		if msg == None:
			return False
		if self._IsError(msg):
			return False
		res = self._SplitString(msg)
		if res[0].upper() != 'STATUS':
			return False
		if res[2].upper() != 'ACCEPTED':
			return False
		return True
		
	def reset(self):
		"""Close any open connection and reset to default state"""
		self.Close()
		self._Username = None
		self._Password = None
		self._IPAddress = None
		self._Port = 2553
		self._Timeout = 4000
		self._LastState = DoorStatus.Unknown
		self._LastAge = 0
		self._LastTemperature = -999.9
		self._LastAlarm = False
		self._AlarmTrigger = AlarmTrigger.NoAlarm
		self._AlarmMessage = None
		self._LastEvents = None

	###############################################################################
	#							"Public" Methods					              #
	###############################################################################

	#
	#	Connection Methods
	#			
		
	def StartUpdateClose(self, numberEvents=0):
		"""Establish secure connection, get latest data, close connection.
		
		If numberEvents > 0 then upto numberEvents will be loaded in stored in
		_LastEvents.  _LastEvents may be None if there were no events
		See Connect(), GetStatus(True), GetEvents(), Close().
		Returns None on error or results of GetStatus()"""
		rc = None
		if self.Connect() > 0:
			rc = self.GetStatus(True)
			if rc != None and numberEvents > 0:
				self._LastEvents = self.GetEvents(numberEvents)
		self.Close()
		return rc
		
	def StartPushClose(self):
		"""Establish secure connection, push button, close connection.
		
		The status of the door will not be updated on this call.  It may take many 
		seconds for the door state to stabilize.  So wait a bit before calling 
		StartUpClose.
		Returns True on success (button push simulated), False otherwise"""
		rc = False
		if self.Connect() > 0:
			rc = self._PushButton()
		self.Close()
		return rc	
			
	def StartClearAlarmClose(self):
		"""Establish secure connection, clear alarm, close connection.
		
		Any active alarm will be cleared.
		Returns True on success (alarm cleared simulated), False otherwise"""
		rc = False
		if self.Connect() > 0:
			rc = self._ClearAlarm()
		self.Close()
		return rc	
		
	def Connect(self):
		"""Establish secure connection to device and login.
		
		username, password, address and optional port must all be initialized.
		The miDoor device supports two active connections so the connections should
		not remain on for too long of a period.
		Connecting is expensive because of SSL encryption
		True if returned if the device was attached and logged-in successfully.  
		-1 for connection error, -2 for authentication error.  1 for successfully
		connecting and logging-in."""
		try:
			if __stdssl__ == True:
				s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)	
				self._Socket = ssl.wrap_socket(s)
				self._Socket.connect((self._IPAddress, self._Port))
				self._Socket.settimeout(self._Timeout/1000)
			else:
				res = socket.getaddrinfo(self._IPAddress, self._Port, 0, socket.SOCK_STREAM)[0]
				s = socket.socket(*res[0:3])
				self._Socket = s
				self._Socket.settimeout(self._Timeout/1000)
				self._SSLConnect(res[4])
		except:
			print "Connect Exception: " + str(sys.exc_info()[0]), sys.exc_info()
			return -1
		if self._Send("LOGON " + self._Username + " " + self._Password + "\n") == False:
			self.Close()
			return -1
		result = self._Read()
		if result == None:
			return -2
		if result == 'CONFIRMED\n':
			return 1
		return -2
		
	def Close(self):
		"""Logout from the device and close the connection"""
		if self._Socket != None:
			try:
				self._Send("EXIT\n")
				if __stdssl__ == False:
					if self._SSLConnection != None:
						self._SSLConnection.shutdown();
						self._SSLConnection.close();		
				self._Socket.close()
			except:
				print "Close Exception: ", sys.exc_info()
				pass
			self._Socket = None
		return

	#
	#	Accessor Methods (miDoor data)
	#
	
	def PushButton(self):
		"""Cause a push operation to be sent to the device.  A valid connection must
		be active.
		
		Returns True if successful, False otherwise"""
		return self._PushButton()
		
	def ClearAlarm(self):
		"""Send Clear Alarm command to the device.  A valid connection must e active.
		
		Returns True if successful, False otherwise"""
		return self._ClearAlarm()
		
	def GetStatus(self, forceUpdate=False):
		"""Get an array of door status.  
		
		If forceUpdate is True then the device is queried otherwise the cached data
		is used.  
		Returns None on error or an array of:
			[<state>, <age>, <temperature celsius>, <alarm>:True/False> ,[ <alarm trigger>, <alarm message>]]"""
		if forceUpdate:
			if self._UpdateDoorStatus() == False:
				return None
		res = [self._LastState, self._LastAge, self._LastTemperature, self._LastAlarm]
		if self._LastAlarm:
			ares = [self._AlarmTrigger, self._AlarmMessage]
			res.append(ares)
		return res;

	def GetEvents(self, numberEvents=256):
		"""Read events into a sorted array
		
		The returned array is an array of events which are themselves arrays.  
		The returned array is ordered so that the most recent is at the top of the
		array and the oldest is the last element in the array.
		Up to numberEvents will be read even if there are more events in the device.
		If no events are returned then None is returned"""
		return self._ReadAllEvents(numberEvents)

	#
	#	Accessor Methods (Connection Settings)
	#
	
	def GetDoorStatus(self):
		"""Return last door status update (DoorStatus)"""
		return self._LastState
		
	def GetTemperature(self):
		"""Return last temperature read form device. -999.9 indicates unknown"""
		return self._LastTemperature
		
	def GetAlarmed(self):
		"""Return last alarm status (AlarmTrigger) read from device."""
		return self._AlarmTrigger
		
	def GetAlarmMessage(self):
		"""Returns the active alarm message (String) or None if there is no alarm"""
		if self._AlarmTrigger < AlarmTrigger.Reserved:
			return None
		return self._AlarmMessage

	def SetAddress(self, ipaddress):
		"""Set the remote address of the device.  Can be ip address or domain name.
		
		The username is not set if there is an active connection.
		True on success."""
		if self._CheckConnection():
			return False
		self._IPAddress = ipaddress
		return True
	def GetAddress(self):
		"""Get the current remote address"""
		return self._IPAddress

	def SetUsername(self, username):
		"""Set the username of the device.
		
		The username is not set if there is an active connection.
		True on success."""
		if self.__CheckConnection():
			return False
		self._Username = username
		return True
	def GetUsername(self):
		"""Get the current username"""
		return self._Username
					
	def SetPassword(self, password):
		"""Set the password of the device.
		
		The password is not set if there is an active connection.
		True on success."""
		if self._CheckConnection():
			return False
		self._Password = password
		return True
	def GetPassword(self):
		"""Get the current password"""
		return self._Password

	def SetPort(self, port):
		"""Set the port of the device.
		
		The port is not set if there is an active connection.
		True on success."""
		if self._CheckConnection():
			return False
		self._Port = port
		return True
	def GetPort(self):
		"""Get the current port"""
		return self._Port

	def SetTimeout(self, timeout=4000):
		"""Set the connection timeout in milliseconds.  
		
		If there is an active connection then the connection is updated"""
		self._Timeout = timeout
		if self._CheckConnection():
			try:
				if __stdssl__ == False and self._SSLConnection != None:
					s = long(self._Timeout/1000)
					us = long((self._Timeout-(s*1000))*1000)
					tv = struct.pack('ll', s, us)
					self._SSLConnection.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, tv)
				else:
					self._Socket.settimeout(self._Timeout/1000)
			except:
				print "SetTimeout Exception: ", sys.exc_info()
				pass
		return
	def GetTimeout(self):
		"""Get the current timeout (ms)"""
		return self._Timeout
	
		
if __name__ == "__main__":
	if __stdssl__ == True:
		print "Using SSL module..."
	else:
		print "Using OpenSSL module..."
	a = miDoor(ipAddress='10.0.1.118', port=2553, user='testuser', password='testpass')
	#print a.Connect()
	#a.SetTimeout(4000)
	#print a._UpdateDoorStatus()
	#print a._LastState, a._LastAge, a._LastTemperature, a._LastAlarm
	#print a.GetStatus(True)
	#print a._UpdateAlarmStatus()
	#a.Close()
	print a.StartUpdateClose(5)
	print a._LastEvents
	print a.StartPushClose()